#!/usr/bin/perl
#--preamble, use declarations
use strict;
use warnings;

use diagnostics;
use IO::Handle;
#--end preamble
# 2004 Timo Latvala <timo.latvala@hut.fi>
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


##subroutine for generating an integer between min and max

sub random_int_in ($$) {
    my($min, $max) = @_;
    # Assumes that the two arguments are integers themselves!
    return $min if $min == $max;
    ($min, $max) = ($max, $min)  if  $min > $max;
    return $min + int rand(1 + $max - $min);
}  


#print the usage of the program

sub usage() {
    print ("prefixToInt [-o filename] [inputfile]\n");
}

#return TRUE when the arguement is t and FALSE when 
#the argument is f

sub tf($) {
    if($_[0] eq "t") {
	return "TRUE";
    }
    elsif($_[0] eq "f") {
	return "FALSE";
    }
    else {
	return "UNDEF";
    }
}
 

#declaration
sub ctlconvert(\@);
#convert an LTL formula in prefix notation to CTL infix form 

sub ctlconvert(\@) {
#    print "ARGLIST: $_[0]";
    my $listref=shift @_;
    my $op=shift @$listref;
#    print "TEST: $op\n";

    if ($op =~ /[XFG]/) {
	return "A$op ".ctlconvert(@$listref);
    }
    elsif ($op =~ /[!]/) {
	return "$op ".ctlconvert(@$listref);
    }
    elsif($op =~ /[&|]/) {
	my $leftOp=ctlconvert(@$listref);
	return "($leftOp $op ".ctlconvert(@$listref).")";
    }
    elsif($op =~ /[UV]/) {
	my $leftOp=ctlconvert(@$listref);
	return "A[$leftOp $op ".ctlconvert(@$listref)."]";
    }
    elsif($op =~ /p\d+/) {
	return $op;
    }
    elsif($op =~ /[tf]/) {
	return tf($op);
    }
    else {
	die "Internal error. No op: $op\n";
    }
}

 
#declaration
sub convert(\@);
#convert an LTL formula in prefix notation to infix form and 
#randomly insert past operators.

sub convert(\@) {
#    print "ARGLIST: $_[0]";
    my $listref=shift @_;
    my $op=shift @$listref;
#    print "TEST: $op\n";

    if ($op =~ /[!XFG]/) {
	if(random_int_in(0,1)==1) {
	    if($op=~/[G]/) {
		return "H ".convert(@$listref);
	    }
	    elsif ($op=~/[F]/) {
		return "O ".convert(@$listref);
	    }
	    elsif ($op=~/[X]/) {
		if(random_int_in(0,1)==1) {
		    return "Y ".convert(@$listref);
		}
		return "Z ".convert(@$listref);
	    }  
	}
	return "$op ".convert(@$listref);
    }
    elsif($op =~ /[UV&|]/) {
	my $leftOp=convert(@$listref);
	return "($leftOp $op ".convert(@$listref).")";
    }
    elsif ($op =~ /[MW]/) {
	my $leftOp=convert(@$listref);
	if ($op =~ /[M]/) {
	    return "($leftOp S ".convert(@$listref).")";
	}
	else {
	    return "($leftOp T ".convert(@$listref).")";
	}
    }
    elsif($op =~ /p\d+/) {
	return $op;
    }
    elsif($op =~ /[tf]/) {
	return tf($op);
    }
    else {
	die "Internal error. No op: $op\n";
    }
}

#---------------------------------------------------------------------
#                      MAIN PROGRAM
#---------------------------------------------------------------------
# Check arguments, open files, and call generator code
#
#---------------------------------------------------------------------

if (@ARGV >4) {
    print STDERR "Too many arguments.\n";
    usage();
    exit 1;
}


#declare parameter variables
my $input_file="";
my $output_file="";
my $ctl="";

for (my $i=0; $i <= $#ARGV; $i++) {
    if($ARGV[$i] eq "-o") {
	$output_file=$ARGV[$i+1];
	($output_file =~ /.+/) or die "Illegal file name specified using -o.\n";
	$i=$i+1;
    }
    elsif($ARGV[$i] eq "-ctl") {
	$ctl="yes";
    }
    else {
	$input_file=$ARGV[$i];
	($input_file =~ /.+/) or die "Illegal input file name specified.\n";	
    }
}



my $IFILE=IO::Handle->new();
my $OFILE=IO::Handle->new();

if ($output_file ne "") { 
    open($OFILE,"> $output_file") or die "Can't open file $output_file: $!";    
} else {
    $OFILE=*STDOUT;
}


if ($input_file ne "") { 
    open($IFILE,"< $input_file") or die "Can't open file $input_file: $!";    
} else {
    $IFILE=*STDIN;
}

#read one line at a time and assign it to $_

if ($ctl eq "yes") {
    while(<$IFILE>) {
	my @formula=split(/\s+/, $_);
	print $OFILE "SPEC ";
	print $OFILE ctlconvert(@formula)."\n";
    }
}
else {
    while(<$IFILE>) {
	my @formula=split(/\s+/, $_);
	print $OFILE "LTLSPEC ";
	print $OFILE convert(@formula)."\n";
    }
}

close $IFILE;
close $OFILE

